/* $Id: tif_unix.c,v 1.8 2004/11/28 15:04:54 dron Exp $ */

/*
 * Copyright (c) 1988-1997 Sam Leffler
 * Copyright (c) 1991-1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

/*
 * Giuseppe Costagliola 2005
 * Modified version for iSeries/AS400
 *
 */

/*
 * TIFF Library UNIX-specific Routines. These are should also work with the
 * Windows Common RunTime Library.
 */
#include "tif_config.h"

#include <stdlib.h>
#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_IO_H
# include <io.h>
#endif

#include "tiffiop.h"

#ifdef AS400
#include <qusec.h>
#include <qmhsndpm.h>
#endif

#ifdef AS400
/* ebcdic-to-ascii */
void ebcdic2ascii(void *dest, const void *srce, size_t count);
#endif

static tsize_t
_tiffReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	return ((tsize_t) read((int) fd, buf, (size_t) size));
}

static tsize_t
_tiffWriteProc(thandle_t fd, tdata_t buf, tsize_t size)
{
#ifndef AS400
	return ((tsize_t) write((int) fd, buf, (size_t) size));
#else
        tdata_t *asciibuf;
        tsize_t written = 0;
        if (size > 0) {
           asciibuf = (tdata_t*)_TIFFmalloc(size *sizeof(tdata_t));
           ebcdic2ascii(asciibuf, buf, size);
           written = ((tsize_t) write((int) fd, asciibuf, (size_t) size));
           _TIFFfree(asciibuf);
           return written;
        }
        if (size < 0) {
           size = -size;
           return ((tsize_t) write((int) fd, buf, (size_t) size));
        }
#endif
}

static toff_t
_tiffSeekProc(thandle_t fd, toff_t off, int whence)
{
	return ((toff_t) lseek((int) fd, (off_t) off, whence));
}

static int
_tiffCloseProc(thandle_t fd)
{
	return (close((int) fd));
}


static toff_t
_tiffSizeProc(thandle_t fd)
{
#ifdef _AM29K
	long fsize;
	return ((fsize = lseek((int) fd, 0, SEEK_END)) < 0 ? 0 : fsize);
#else
	struct stat sb;
	return (toff_t) (fstat((int) fd, &sb) < 0 ? 0 : sb.st_size);
#endif
}

#ifdef HAVE_MMAP
#include <sys/mman.h>

static int
_tiffMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
	toff_t size = _tiffSizeProc(fd);
	if (size != (toff_t) -1) {
		*pbase = (tdata_t)
		    mmap(0, size, PROT_READ, MAP_SHARED, (int) fd, 0);
		if (*pbase != (tdata_t) -1) {
			*psize = size;
			return (1);
		}
	}
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
	(void) fd;
	(void) munmap(base, (off_t) size);
}
#else /* !HAVE_MMAP */
static int
_tiffMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
	(void) fd; (void) pbase; (void) psize;
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
	(void) fd; (void) base; (void) size;
}
#endif /* !HAVE_MMAP */

/*
 * Open a TIFF file descriptor for read/writing.
 */
TIFF*
TIFFFdOpen(int fd, const char* name, const char* mode)
{
	TIFF* tif;

	tif = TIFFClientOpen(name, mode,
	    (thandle_t) fd,
	    _tiffReadProc, _tiffWriteProc,
	    _tiffSeekProc, _tiffCloseProc, _tiffSizeProc,
	    _tiffMapProc, _tiffUnmapProc);
	if (tif)
		tif->tif_fd = fd;
	return (tif);
}

/*
 * Open a TIFF file for read/writing.
 */
TIFF*
TIFFOpen(const char* name, const char* mode)
{
	static const char module[] = "TIFFOpen";
	int m, fd;
        TIFF* tif;

	m = _TIFFgetMode(mode, module);
	if (m == -1)
		return ((TIFF*)0);

/* for cygwin and mingw */
#ifdef O_BINARY
        m |= O_BINARY;
#endif

#ifdef _AM29K
	fd = open(name, m);
#else
#ifdef AS400
	fd = open(name, m | O_CODEPAGE, 0666, 850);
#else
	fd = open(name, m, 0666);
#endif
#endif

	if (fd < 0) {
		TIFFError(module, "%s: Cannot open", name);
		return ((TIFF *)0);
	}

	tif = TIFFFdOpen((int)fd, name, mode);
	if(!tif)
		close(fd);
	return tif;
}

#ifdef __WIN32__
/*
 * Open a TIFF file with a Unicode filename, for read/writing.
 */
TIFF*
TIFFOpenW(const wchar_t* name, const char* mode)
{
	static const char module[] = "TIFFOpenW";
	int m, fd;
	int mbsize;
	char *mbname;
	TIFF* tif;

	m = _TIFFgetMode(mode, module);
	if (m == -1)
		return ((TIFF*)0);

/* for cygwin and mingw */
#ifdef O_BINARY
        m |= O_BINARY;
#endif

	fd = _wopen(name, m, 0666);
	if (fd < 0) {
		TIFFError(module, "%s: Cannot open", name);
		return ((TIFF *)0);
	}

	mbname = NULL;
	mbsize = WideCharToMultiByte(CP_ACP, 0, name, -1, NULL, 0, NULL, NULL);
	if (mbsize > 0) {
		mbname = _TIFFmalloc(mbsize);
		if (!mbname) {
			TIFFError(module,
			"Can't allocate space for filename conversion buffer");
			return ((TIFF*)0);
		}

		WideCharToMultiByte(CP_ACP, 0, name, -1, mbname, mbsize,
				    NULL, NULL);
	}

	tif = TIFFFdOpen((int)fd, (mbname != NULL) ? mbname : "<unknown>",
			 mode);
	
	_TIFFfree(mbname);
	
	if(!tif)
		close(fd);
	return tif;
}
#endif

void*
_TIFFmalloc(tsize_t s)
{
	return (malloc((size_t) s));
}

void
_TIFFfree(tdata_t p)
{
	free(p);
}

void*
_TIFFrealloc(tdata_t p, tsize_t s)
{
	return (realloc(p, (size_t) s));
}

void
_TIFFmemset(tdata_t p, int v, tsize_t c)
{
	memset(p, v, (size_t) c);
}

void
_TIFFmemcpy(tdata_t d, const tdata_t s, tsize_t c)
{
	memcpy(d, s, (size_t) c);
}

int
_TIFFmemcmp(const tdata_t p1, const tdata_t p2, tsize_t c)
{
	return (memcmp(p1, p2, (size_t) c));
}

static void
unixWarningHandler(const char* module, const char* fmt, va_list ap)
{
#ifndef AS400
	if (module != NULL)
		fprintf(stderr, "%s: ", module);
	fprintf(stderr, "Warning, ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, ".\n");
#else
        int rc = 0;
        char error_module[64];
        char error_text[512];
        Qus_EC_t   err_code;
        char       msg_key [4];
        char       *msg_dta;

        if (module != NULL)
          rc = sprintf(error_module, "%s: ", module);
        else
          strcpy(error_module, "tiff2pdf");
        rc = vsprintf(error_text, fmt, ap);

        msg_dta = (char*)_TIFFmalloc((strlen(error_module) +
                                      strlen(error_text) + 1)
                                      *sizeof(char));
        strcpy(msg_dta, error_module);
        strcat(msg_dta, error_text);

        /* Send diag message */
        err_code.Bytes_Provided = 32;
        QMHSNDPM (
           "CPF9898",                /* message id                  */
           "QCPFMSG   *LIBL     ",   /* Qualified message file name */
           msg_dta,                  /* Message data                */
           strlen(msg_dta),          /* Length of message data      */
           "*DIAG     ",             /* Message type                */
           "*PGMBDY   ",             /* Call stack entry            */
           1,                        /* Call stack counter          */
           msg_key,                  /* Message key                 */
           &err_code                 /* Error code                  */
        );
        _TIFFfree(msg_dta);
#endif
}
TIFFErrorHandler _TIFFwarningHandler = unixWarningHandler;

static void
unixErrorHandler(const char* module, const char* fmt, va_list ap)
{
#ifndef AS400
	if (module != NULL)
		fprintf(stderr, "%s: ", module);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, ".\n");
#else
        int rc = 0;
        char error_module[64];
        char error_text[512];
        Qus_EC_t   err_code;
        char       msg_key [4];
        char       *msg_dta;
        if (module != NULL)
          rc = sprintf(error_module, "%s: ", module);
        else
          strcpy(error_module, "tiff2pdf");
        rc = vsprintf(error_text, fmt, ap);
        msg_dta = (char*)_TIFFmalloc((strlen(error_module) +
                                      strlen(error_text) + 1)
                                      *sizeof(char));
        strcpy(msg_dta, error_module);
        strcat(msg_dta, error_text);

        /* Send escape message */
        err_code.Bytes_Provided = 32;
        QMHSNDPM (
           "CPF9898",                /* message id                  */
           "QCPFMSG   *LIBL     ",   /* Qualified message file name */
           msg_dta,                  /* Message data                */
           strlen(msg_dta),          /* Length of message data      */
           "*ESCAPE   ",             /* Message type                */
           "*PGMBDY   ",             /* Call stack entry            */
           1,                        /* Call stack counter          */
           msg_key,                  /* Message key                 */
           &err_code                 /* Error code                  */
        );
        _TIFFfree(msg_dta);
#endif
}
TIFFErrorHandler _TIFFerrorHandler = unixErrorHandler;

#ifdef AS400
void
ebcdic2ascii(void *dest, const void *srce, size_t count)
{

  static const unsigned char tbl_toascii[256] = {
      0x00,0x01,0x02,0x03,0xDC,0x09,0xC3,0x1C,
      0xCA,0xB2,0xD5,0x0B,0x0C,0x0D,0x0E,0x0F,
      0x10,0x11,0x12,0x13,0xDB,0xDA,0x08,0xC1,
      0x18,0x19,0xC8,0xF2,0x1A,0x1D,0x1E,0x1F,
      0xC4,0xB3,0xC0,0xD9,0xBF,0x0A,0x17,0x1B,
      0xB4,0xC2,0xC5,0xB0,0xB1,0x05,0x06,0x07,
      0xCD,0xBA,0x16,0xBC,0xBB,0xC9,0xCC,0x04,
      0xB9,0xCB,0xCE,0xDF,0x14,0x15,0xFE,0x7F,
      0x20,0xFF,0x83,0x84,0x85,0xA0,0xC6,0x86,
      0x87,0xA4,0xBD,0x2E,0x3C,0x28,0x2B,0x7C,
      0x26,0x82,0x88,0x89,0x8A,0xA1,0x8C,0x8B,
      0x8D,0xE1,0x21,0x24,0x2A,0x29,0x3B,0xAA,
      0x2D,0x2F,0xB6,0x8E,0xB7,0xB5,0xC7,0x8F,
      0x80,0xA5,0xDD,0x2C,0x25,0x5F,0x3E,0x3F,
      0x9B,0x90,0xD2,0xD3,0xD4,0xD6,0xD7,0xD8,
      0xDE,0x60,0x3A,0x23,0x40,0x27,0x3D,0x22,
      0x9D,0x61,0x62,0x63,0x64,0x65,0x66,0x67,
      0x68,0x69,0xAE,0xAF,0xD0,0xEC,0xE7,0xF1,
      0xF8,0x6A,0x6B,0x6C,0x6D,0x6E,0x6F,0x70,
      0x71,0x72,0xA6,0xA7,0x91,0xF7,0x92,0xCF,
      0xE6,0x7E,0x73,0x74,0x75,0x76,0x77,0x78,
      0x79,0x7A,0xAD,0xA8,0xD1,0xED,0xE8,0xA9,
      0x5E,0x9C,0xBE,0xFA,0xB8,0xF5,0xF4,0xAC,
      0xAB,0xF3,0x5B,0x5D,0xEE,0xF9,0xEF,0x9E,
      0x7B,0x41,0x42,0x43,0x44,0x45,0x46,0x47,
      0x48,0x49,0xF0,0x93,0x94,0x95,0xA2,0xE4,
      0x7D,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,0x50,
      0x51,0x52,0xFB,0x96,0x81,0x97,0xA3,0x98,
      0x5C,0xF6,0x53,0x54,0x55,0x56,0x57,0x58,
      0x59,0x5A,0xFD,0xE2,0x99,0xE3,0xE0,0xE5,
      0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,
      0x38,0x39,0xFC,0xEA,0x9A,0xEB,0xE9,0x9F
  };
        unsigned char *udest = dest;
        const unsigned char *usrce = srce;

        while (count-- != 0) {
            *udest++ = tbl_toascii[*usrce++];
        }

}
#endif
